Require Import Nat Bool List.

From PCF Require Export Terms.

(** * Values *)

Inductive value : form -> Prop :=
  | v_abs : 
      forall t,
        is_closed (f_abs t) ->
        value (f_abs t)

  | v_star : 
      value (f_star)
  | v_nat : 
      forall n, 
        value (f_nat n)

  | v_pair : forall t s, 
      is_closed t ->
      is_closed s ->
      value (f_pair t s)

  | v_inl : forall t,
      is_closed t ->
      value (f_inl t)
  | v_inr : forall t,
      is_closed t ->
      value (f_inr t).

(** * Big-Step Call-by-Name *)

Inductive bigCBN : form -> form -> Prop :=
  | bigCBN_value :
      forall v,
        value v ->
        bigCBN v v
  (** Beta reduction *)
  | bigCBN_app : 
      forall p p' q r, 
        bigCBN p (f_abs p') ->
        bigCBN (subst p' 0 q) r -> 
        bigCBN (f_app p q) r

  | bigCBN_fst : 
      forall t p q r, 
        bigCBN t (f_pair p q) ->
        bigCBN p r ->
        bigCBN (f_fst t) r
  | bigCBN_snd :
      forall t p q r, 
        bigCBN t (f_pair p q) ->
        bigCBN q r ->
        bigCBN (f_snd t) r

  | bigCBN_case_inl :
      forall b b' p r q,
        bigCBN b (f_inl b') ->
        bigCBN (subst p 0 b') r ->
        bigCBN (f_case b p q) r
  | bigCBN_case_inr :
      forall b b' p r q,
        bigCBN b (f_inr b') ->
        bigCBN (subst q 0 b') r ->
        bigCBN (f_case b p q) r

  | bigCBN_fix : 
      forall f A r, 
        bigCBN (f_app f (f_app (f_fix A) f)) r ->
        bigCBN (f_app (f_fix A) f) r

        (** I wanted to remove the following two rules, 
        but I wasn't able to proove them with the other rules. In addition, 
        I'm currently not convinced, that they are solvable because of binding issues. *)
  | bigCBN_ite_true :
      forall b p r q, 
        bigCBN b (f_bool true) ->
        bigCBN p r ->
        bigCBN (f_ite b p q) r
  | bigCBN_ite_false :
      forall b p r q, 
        bigCBN b (f_bool false) ->
        bigCBN q r ->
        bigCBN (f_ite b p q) r

  | bigCBN_le : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = (n1 <=? n2) ->
        bigCBN (f_le p q) (f_bool m)
  | bigCBN_lt : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = (n1 <? n2) ->
        bigCBN (f_lt p q) (f_bool m)

  | bigCBN_add : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = n1 + n2 ->
        bigCBN (f_add p q) (f_nat m)
  | bigCBN_sub : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = n1 - n2 ->
        bigCBN (f_sub p q) (f_nat m)
  | bigCBN_mul : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = n1 * n2 ->
        bigCBN (f_mul p q) (f_nat m)
  | bigCBN_div : 
      forall n1 n2 m p q, 
        bigCBN p (f_nat n1) ->
        bigCBN q (f_nat n2) ->
        m = n1 / n2 ->
        bigCBN (f_div p q) (f_nat m).

Theorem bigCBN_reduces_to_value :
  forall p v,
    bigCBN p v ->
    value v.
Proof.
  induction 1 as
  [v H1 (* bigCBN_value *)
  |p p' q r H1 IH1 H2 IH2 (* bigCBN_app *)

  |t p q r H1 IH1 H2 IH2 (* bigCBN_fst *)
  |t p q r H1 IH1 H2 IH2 (* bigCBN_snd *)

  |b b' p r q H1 IH1 H2 IH2 (* bigCBN_case_inl *)
  |b b' p r q H1 IH1 H2 IH2 (* bigCBN_case_inr *)

  |f A r H1 IH1 (* bigCBN_fix *)

  |b p r q H1 IH1 H2 IH2 (* bigCBN_ite_true *)
  |b p r q H1 IH1 H2 IH2 (* bigCBN_ite_false *)

  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_le *)
  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_lt *)

  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_add *)
  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_sub *)
  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_mul *)
  |n1 n2 m p q H1 IH1 H2 IH2 H3 (* bigCBN_div *)
  ];
  eauto using bigCBN, value; destruct m; constructor; reflexivity.
Qed.

Theorem bigCBN_bool :
  forall b, 
    bigCBN (f_bool b) (f_bool b).
Proof.
  intros [|]; do 2 constructor; reflexivity.
Qed.
